@@ -61,7 +61,7 @@ static char        *orbit_debug_options      = NULL;
 static char        *orbit_naming_ref         = NULL;
 static GSList      *orbit_initref_list       = NULL; 
 static gboolean     orbit_use_corbaloc       = FALSE;
-static gint         orbit_timeout_limit      = -1;
+static guint        orbit_timeout_msec       = 60000; /* 60 seconds - 0 will disable timeouts altogether */
 void
 ORBit_ORB_start_servers (CORBA_ORB orb)
 {
@@ -418,7 +418,7 @@ CORBA_ORB_init (int *argc, char **argv,
 #endif /* G_ENABLE_DEBUG */
 
 	giop_recv_set_limit (orbit_initial_recv_limit);
-	giop_recv_set_timeout (orbit_timeout_limit);
+	giop_set_timeout (orbit_timeout_msec);
 	giop_init (thread_safe,
 		   orbit_use_ipv4 || orbit_use_ipv6 ||
 		   orbit_use_irda || orbit_use_ssl);
@@ -1323,6 +1323,7 @@ CORBA_ORB_destroy (CORBA_ORB          orb,
 
 	/* destroy the forward bind hashtable*/
 	g_hash_table_destroy (orb->forw_binds);
+	orb->forw_binds = NULL;
 	
 	orb->life_flags |= ORBit_LifeF_Destroyed;
 
@@ -1467,8 +1468,8 @@ const ORBit_option orbit_supported_options[] = {
 	{ "ORBDebugFlags",      ORBIT_OPTION_STRING,  &orbit_debug_options },
 	{ "ORBInitRef",         ORBIT_OPTION_KEY_VALUE,  &orbit_initref_list},
 	{ "ORBCorbaloc",        ORBIT_OPTION_BOOLEAN, &orbit_use_corbaloc},
-	{ "GIOPTimeoutLimit",   ORBIT_OPTION_INT,     &orbit_timeout_limit },
-	{ NULL,                 0,                    NULL },
+	{ "GIOPTimeoutMSEC",    ORBIT_OPTION_ULONG,   &orbit_timeout_msec },
+	{ NULL,                 0,                    NULL }
 };
 
 #ifdef G_OS_WIN32
